/*
############################################################################
# Lock test program
# Copyright (c) 2015 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <propeller.h>

// Define either USE_PROP_LOCK or USE_MUTEX_LOCK
#define USE_PROP_LOCK
#undef  USE_MUTEX_LOCK

// Check if both locks are defined
#if defined(USE_PROP_LOCK) && defined(USE_MUTEX_LOCK)
#error Both USE_PROP_LOCK and USE_MUTEX_LOCK are defined
#error Define only one of them
#endif

// Set the number of pthreads between 0 and 6
#define NUM_PTHREADS 2

// Stack size of 1000 bytes should be more than enough
#define PTHREAD_STACKSIZE 1000

#ifdef USE_PROP_LOCK
int prop_lock;
#endif
#ifdef USE_MUTEX_LOCK
pthread_mutex_t mutex_lock;
#endif

static int loopcount[8];
pthread_t threads[NUM_PTHREADS];
int stacks[NUM_PTHREADS][PTHREAD_STACKSIZE/4];
static volatile int active_thread = -1;

void *ThreadFunc(void *arg);
void StartPthreads(void);

int main()
{ 
    sleep(1);
    printf("Lock Test\n");
#if NUM_PTHREADS
    StartPthreads();
#endif
    ThreadFunc(0);
    return 0;
}

int RandVal()
{
    return (CNT & 0x7fffffff)%5000 + 5000;
}

static void TestLock(int instance)
{
#ifdef USE_PROP_LOCK
    while (lockset(prop_lock));
#endif
#ifdef USE_MUTEX_LOCK
    pthread_mutex_lock(&mutex_lock);
#endif
    if (active_thread != -1)
        printf("Error: Thread %d, Thread %d also active\n", instance, active_thread);
    active_thread = instance;
    usleep(RandVal());
    active_thread = -1;
#ifdef USE_PROP_LOCK
    lockclr(prop_lock);
#endif
#ifdef USE_MUTEX_LOCK
    pthread_mutex_unlock(&mutex_lock);
#endif
}

// This routine is run by all the pthreads and by the main thead
void *ThreadFunc(void *arg)
{
    int instance = (int)arg;
    
    loopcount[instance] = 0;
    
    printf("Starting thread %d\n", instance);

    while (1)
    {
        usleep(RandVal());
        TestLock(instance);
        if (!((++loopcount[instance]) % 100))
            printf("Thread %d, Loop %d\n", instance, loopcount[instance]);
    }

    return 0;
}

#if NUM_PTHREADS
void StartPthreads(void)
{
    int i;
    pthread_attr_t attr;

#ifdef USE_PROP_LOCK    
    prop_lock = locknew();
#endif
#ifdef USE_MUTEX_LOCK
    pthread_mutex_init(&mutex_lock, NULL);
#endif

    for (i = 0; i < NUM_PTHREADS; i++)
    {
        printf("StartPthreads: %d\n", i+1);
        pthread_attr_init(&attr);
        pthread_attr_setstacksize(&attr, PTHREAD_STACKSIZE);
        pthread_attr_setstackaddr(&attr, stacks[i]);

        if (pthread_create(&threads[i], &attr, ThreadFunc, (void *)i+1))
        {
            printf("pthread_create %d failed\n", i);
            exit(1);
        }
    }
}
#endif

/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
